// Unit ParticleEngine
(**************************************************

- Enthlt: TParticleEmitter

**************************************************)

unit F2D_ParticleEngine;

interface

uses
  windows,
  dglOpenGL,
  graphics,
  F2D_Vectors,
  F2D_Math,
  F2D_Colors,
  F2D_Core,
  F2D_Const;

type
  //Keyframe, fr den Partikel Status am Anfang und am Ende der Simulation
  TKeyFrame = record
    velocity: integer;    //Der "Betrag" der geschwindigkeit
    alpha: integer;       //Alpha wert
    size: TVector2D;      //Gre
    Color: TColor;        //Farbe
  end;

  TParticle = record
    dir,                  //Jedes Partikel bekommt seine eigene Richtung in
                          //die es sich mit der errechneten Geschwindigkeit bewegt
    position: TVector2D;  //Die aktuelle Position des Partikels
    age: single;          //Das Alter des Partikels
  end;

  TParticleEmitter = class
    constructor Create(pEngine: TEngine);
    procedure Free;
  private
    TotalEmissionTime: single;
    FirstTime: boolean;
    FEngine: TEngine;                                //Pointer auf die Engine
    FAngle,                                          //Emissionsrichtung
    FEmissionAngle: integer;                         //Emissionsbereich 0-360
    FParticle: TParticle;                            //Template Partikel
    FParticles: array of TParticle;                  //Partikel Array
    EmissionTime: single;                            //Emissionszeit
    FLifeSpan: integer;                              //Lebensdauer eines Partikels
    function GetParticleCount: integer;              //Anzahl der Partikel
    procedure Add(pParticle: TParticle);             //Partikel hinzufgen
    procedure SetLifeSpan(const Value: integer);     //
    procedure CleanUp;                               //Aufrumen, tote Partikel lschen
    procedure SetAngle(const Value: integer);        //
    procedure SetEmissionAngle(const Value: integer);//
    procedure NextStep;                              //Berechnet den nchsten Schritt
  public
    Duration: integer;                               //Dauer der Emission - wenn 0 dann unendlich
    position: TVector2D;                             //Die Position des Emitters
    EmissionRate,                                    //Emissionsrate (alle x ms)
    ParticlesPerEmission: integer;                   //Partikel pro Emission
    LifeBegin,                                       //Status der Partikel am Anfang und am Ende ihres Lebens
    LifeEnd: TKeyFrame;
    Texture: string;                                 //Textur der Partikel
    z: integer;                                      //Z Tiefe - sollen die Partikel aber ber einem festen Objekt gezeichnet werden,
                                                     //so muss der Aufruf von render auch nach diesem Objekt erfolgen
    property Angle: integer read FAngle write SetAngle;
    property EmissionAngle: integer read FEmissionAngle write SetEmissionAngle;
    property LifeSpan: integer read FLifeSpan write SetLifeSpan;
    property NumParticles: integer read GetParticleCount;

    procedure Render;                                //Zeichnen
  end;

implementation

constructor TParticleEmitter.Create(pEngine: TEngine);
begin
  inherited Create;
  FEngine := pEngine;
  setlength(FParticles, 0);
  FAngle := 0;
  FEmissionAngle := 0;
  EmissionTime := 0;
  FLifeSpan := 0;
  Emissionrate := 0;
  ParticlesPerEmission := 0;
  TotalEmissionTime := 0;
  z := 0;
  Duration := 0;
  FirstTime := true;
end;

function TParticleEmitter.GetParticleCount: integer;
begin
  result := length(FParticles);
end;

procedure TParticleEmitter.SetLifeSpan(const Value: integer);
begin
  if Value > 0 then
    FLifeSpan := Value;
end;

procedure TParticleEmitter.CleanUp;
var
  i, j: integer;
begin
  j := 0;
//Alle tote Partikel aus dem Array entfernen
  for i := 0 to high(FParticles)do
    if FParticles[i].Age < FLifeSpan then
    begin
      FParticles[j] := FParticles[i];
      inc(j);
    end;
  setlength(FParticles,j);
end;

procedure TParticleEmitter.Add(pParticle: TParticle);
begin
//Partikel hinzufgen
  setlength(FParticles, length(FParticles) + 1);
  FParticles[high(FParticles)] := pParticle;
end;

procedure TParticleEmitter.Free;
begin
  if Self <> nil then
    Destroy;
end;

procedure TParticleEmitter.SetAngle(const Value: integer);
begin
  if (Value > 360) then
  begin
    FAngle := Value - 360;
    Exit;
  end;
  if Value < 0 then
  begin
    FAngle := 360 + Value;
    Exit;
  end;

  FAngle := Value;
end;

procedure TParticleEmitter.SetEmissionAngle(const Value: integer);
begin
  if (Value > 360) then
  begin
    FEmissionAngle := Value - 360;
    Exit;
  end;
  if Value < 0 then
  begin
    FEmissionAngle := 360 + Value;
    Exit;
  end;

  FEmissionAngle := Value;
end;

procedure TParticleEmitter.NextStep;
var
  i, c: integer;
  vel: single;
begin
  if Duration > 0 then
    TotalEmissionTime := TotalEmissionTime + 1000 * FEngine.Timestep;
  //Emissionszeit erhhen
  EmissionTime := EmissionTime + 1000 * FEngine.Timestep;
  //Partikel Emittieren
  if (EmissionTime >= EmissionRate) or Firsttime then
  //Neue Partikel erzeugen
  if (TotalEmissionTime <= Duration) or (Duration <= 0) then
    for c := 1 to ParticlesPerEmission do
    begin
      if FirstTime then
        Firsttime := false;
      with FParticle do
      begin
        //Partikel ausrichten
        if EmissionAngle = 360 then
          dir := v2d_normalize(angle_to_v2d(random(360))) else
            dir := v2d_normalize(angle_to_v2d(random(EmissionAngle) + Angle + 90 - EmissionAngle / 2 - 180));
        age := 0;
      end;
      FParticle.position := position;
      Add(FParticle);    
      EmissionTime := 0;
    end;
  //Bewegen
  for i := 0 to high(FParticles) do
    with FParticles[i] do
    begin
      //Partikel altern lassen
      age := age + 1000 * FEngine.Timestep;

      //Geschwindigkeit Interpolieren
      vel := LifeBegin.velocity + (age * (LifeEnd.velocity - LifeBegin.velocity)) / FLifeSpan;

      //neue Position setzen
      position.x := position.x + (dir.x * vel) * FEngine.Timestep;
      position.y := position.y + (dir.y * vel) * FEngine.Timestep;
    end;
  //Aufrumen
  CleanUp;
end;

procedure TParticleEmitter.Render;
var
  i: integer;
  TmpWidth, TmpHeight: Extended;
  size: TVector2D;
  alpha: single;
  color: TColor;
  r,
  g,
  b: integer;
begin
  NextStep;

  for i := 0 to high(FParticles) do
  begin
    glPushMatrix;

    //alpha interpolieren
    alpha := LifeBegin.alpha + (FParticles[i].age * (LifeEnd.alpha - LifeBegin.alpha)) / FLifeSpan;

    // Matrix auf die Zeichenposition setzen
    glTranslatef(FParticles[i].Position.X * (FEngine.Form.ClientWidth / FEngine.Width),
                  FParticles[i].Position.Y * (FEngine.Form.ClientHeight / FEngine.Height), 0);

    FEngine.TextureManager.BindTexture(Texture);

    //Farbe Interpolieren
    r := round(getrvalue(LifeBegin.Color) + (FParticles[i].age * (getrvalue(LifeEnd.Color) - getrvalue(LifeBegin.color))) / FLifeSpan);
    g := round(getgvalue(LifeBegin.Color) + (FParticles[i].age * (getgvalue(LifeEnd.Color) - getgvalue(LifeBegin.color))) / FLifeSpan);
    b := round(getbvalue(LifeBegin.Color) + (FParticles[i].age * (getbvalue(LifeEnd.Color) - getbvalue(LifeBegin.color))) / FLifeSpan);

    color := RGB(r, g, b);

    RGBAf(Color, round(alpha));

    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glEnable(GL_BLEND);

    //Gre interpolieren
    size.x := LifeBegin.size.x + (FParticles[i].age * (LifeEnd.size.x - LifeBegin.size.x)) / FLifeSpan;
    size.y := LifeBegin.size.y + (FParticles[i].age * (LifeEnd.size.y - LifeBegin.size.y)) / FLifeSpan;

    TmpWidth  := (Size.X * (FEngine.Form.ClientWidth  / FEngine.Width))  / 2;
    TmpHeight := (Size.Y * (FEngine.Form.ClientHeight / FEngine.Height)) / 2;

    //Zeichnen
    glBegin(GL_QUADS);
      glTexCoord2f(0,0); glVertex3f(-TmpWidth, -TmpHeight, Z - MaxDepth + 2);
      glTexCoord2f(1,0); glVertex3f( TmpWidth, -TmpHeight, Z - MaxDepth + 2);
      glTexCoord2f(1,1); glVertex3f( TmpWidth,  TmpHeight, Z - MaxDepth + 2);
      glTexCoord2f(0,1); glVertex3f(-TmpWidth,  TmpHeight, Z - MaxDepth + 2);
    glEnd;

    glPopMatrix;
  end;

  glDisable(GL_BLEND);
end;

Initialization
  randomize;

end.
