// Unit Text
(**************************************************

- Enthlt: TFontManager  {--SH 20061111}

**************************************************)
unit F2D_Text;

interface

uses
  // Bentigte Units
  SysUtils, IniFiles, dglOpenGL,
  F2D_BugTracker,F2D_Colors,graphics,Windows;

type
  TFontInfo = class
    ID:TGLuint;
    H:HFont;
    constructor Create(_ID:TGLUint; _H:HFONT);
  end;

  TFontManager = class
    Fonts: THashedStringList;
    // Bugtracker (siehe: UBugTracker)
    BugTracker: TBugTracker;
    FDC:HDC;
    constructor Create(DC:HDC);
    destructor Destroy; override;
  public
    SelectedFont:string;
    // Font hinzufgen
    procedure AddFont(Name, FontName: string; Size: integer; Styles: TFontStyles); overload;
    procedure AddFont(Name: string; Font:TFont; CreateNew:boolean=false); overload;
    function TextSize(Name,Text:string):TSize;
    function GeneralTextHeight(Name:string):single;
    procedure BindFont(pFontName: String; Color:TColor);
  end;

const
  FNT_CHAR_COUNT    =   256;

implementation


{ TFontInfo }

constructor TFontInfo.Create(_ID:TGLUint; _H:HFONT);
begin
  inherited Create;
  ID:= _ID;
  H:= _H;
end;

constructor TFontManager.Create(DC:HDC);
begin
  inherited Create;

  // HashedStringList erstellen
  Fonts := THashedStringList.Create;               

  SelectedFont:= '';
  FDC:= DC;
end;

destructor TFontManager.Destroy;
begin
  // HashedStringList freigeben
  Fonts.Free;

  inherited Destroy;
end;

function InitFontByHandle(DC:HDC; Font:HFont):TFontInfo;
begin
  Result:= TFontInfo.Create(glGenLists(FNT_CHAR_COUNT),0);
  SelectObject(DC, Font);
  wglUseFontBitmaps (DC, 0, FNT_CHAR_COUNT-1, Result.ID);
  Result.H:= Font;
end;

function InitFont(Name:String;DC:HDC; Height:integer;Style:TFontStyles=[]):TFontInfo;
var
  CustomFont:HFont;
  w:integer;
begin
  if fsBold in Style then
    w:= 700
  else
    w:= 0;
  (*CustomFont := GetStockObject (SYSTEM_FONT);*)//Eine Alternative zu CreateFont
  CustomFont := CreateFont(Height,              // Hhe
                           0,                   // Breite 0=Keine Vorgabe
                           0,
                           0,
                           w,                         // Fett?
                           Byte(fsItalic    in Style),// Kursiv?
                           Byte(fsUnderline in Style),// Unterstrichen?
                           Byte(fsStrikeout in Style),// Durchgestrichen?
                           ANSI_CHARSET,
                           OUT_TT_PRECIS,
                           CLIP_DEFAULT_PRECIS,
                           NONANTIALIASED_QUALITY,
                           FF_DONTCARE or DEFAULT_PITCH,
                           PChar(Name));        // Name der Schrift
  Result:= InitFontByHandle(DC,CustomFont);
end;

procedure TFontManager.AddFont(Name, FontName: string; Size: integer; Styles: TFontStyles);
begin
  if (Fonts.IndexOf(name) = -1) then
    begin
      // Array Lnge erhhen
      Size:= MulDiv(Size, GetDeviceCaps(FDC, LOGPIXELSY), 72);

      Fonts.AddObject(Name,InitFont(FontName,FDC,Size,Styles));

      // BugTracker zum Log Datei schreiben veranlassen (O bedeutet erfolgreich)
      BugTracker.Write('O Font: ' + Name);
      BugTracker.Write('');
    end;
end;

procedure TFontManager.AddFont(Name: string; Font: TFont; CreateNew:boolean=false);
begin
  if (not CreateNew)then begin
    if (Fonts.IndexOf(name) = -1) then
    begin
      Fonts.AddObject(Name,InitFontByHandle(FDC,Font.Handle));

      // BugTracker zum Log Datei schreiben veranlassen (O bedeutet erfolgreich)
      BugTracker.Write('O Font: ' + Name);
      BugTracker.Write('');
    end;
  end else
    AddFont(Name,Font.Name,Font.Size,Font.Style);
end;

procedure TFontManager.BindFont(pFontName: String; Color:TColor);
var f:TGLuint;
begin
  if pFontName<>SelectedFont then begin
    if Fonts.IndexOf(pFontName)>=0 then
      f:= TFontInfo(Fonts.Objects[Fonts.IndexOf(pFontName)]).ID
    else
      f:= 0;
    glListBase(f);
    SelectedFont:= pFontName;
  end;

  RGBf(Color);
end;

function TFontManager.TextSize(Name, Text: string): TSize;
var f:HFont;
    old:HGDIOBJ;
begin
  if Fonts.IndexOf(Name)>=0 then
    f:= TFontInfo(Fonts.Objects[Fonts.IndexOf(Name)]).H
  else begin
    Result.cx:= 0;
    Result.cy:= 0;
    exit;
  end;
  old:=  SelectObject(FDC,f);
  GetTextExtentPoint32(FDC,pchar(Text),length(Text),Result);
  SelectObject(FDC,old);
end;

function TFontManager.GeneralTextHeight(Name: string): single;
begin
  Result:= TextSize(name,'1IqQ32').cy;
end;

end.
