// Unit UI
(**************************************************

- Enthlt: TBarGraph
           TListBox

**************************************************)

unit F2D_UI;

interface

uses
  // Bentigte Units
  Graphics, IniFiles, dglOpenGL, glBitmap, Classes,
  F2D_Vectors,
  F2D_Const,
  F2D_Core,
  F2d_Objects,
  F2D_Colors,
  F2D_TextureManager;

type
  TUIElement = class(TSimpleObject)
  protected
    procedure SetMatrix; override;
  end;


  TBarGraph = class(TUIElement)
    constructor Create(pEngine: TEngine);
  protected
    procedure BindTexture;   override;
    procedure DrawInternal(pUseTexture: Boolean); override;
  public
    MaxVal,Value:single;
    TextFG,TextBG:string;
    ColorFG:TColor;
    Horizontal:boolean;
  end;

  TListBox = class(TUIElement)
    constructor Create(pEngine: TEngine);
    destructor Destroy; override;
  protected
    procedure BindTexture;   override;
    procedure DrawInternal(pUseTexture: Boolean); override;
  public
    Items:TStringList;
    ItemIndex:integer;
    Border:single;
    Font:string;
    TextColor,
    SelectedColor:TColor;
    procedure MoveSelection(Delta:integer);
    function SelectedItem:string;
  end;

implementation

uses F2D_Text;


{ TUIElement }

procedure TUIElement.SetMatrix;
begin
  Ofs:= V2D_Scale(Size,-0.5);
  inherited SetMatrix;
end;

{ TBarGraph }

constructor TBarGraph.Create(pEngine: TEngine);
begin
  inherited;
  MaxVal:= 100;
  Value:= 100;
  TextFG:= '';
  TextBG:= '';
  Color:= clGreen;
  ColorFG:= clRed;
  Horizontal:=true;
end;

procedure TBarGraph.BindTexture;
begin
  FEngine.TextureManager.BindTexture(TextBG);
end;

procedure TBarGraph.DrawInternal(pUseTexture: Boolean);
var
  TmpWidth, TmpHeight,perc: Extended;
begin
  // Das eigentlich "Zeichnen"
  TmpWidth  := Size.X * (FEngine.Form.ClientWidth  / FEngine.Width) ;
  TmpHeight := Size.Y * (FEngine.Form.ClientHeight / FEngine.Height);

  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

  glBegin(GL_QUADS);
    glTexCoord2f(0, 0); glVertex3f(0,        0, Z - MaxDepth + 2);
    glTexCoord2f(1, 0); glVertex3f(TmpWidth, 0, Z - MaxDepth + 2);
    glTexCoord2f(1, 1); glVertex3f(TmpWidth,TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f(0, 1); glVertex3f(0,       TmpHeight, Z - MaxDepth + 2);
  glEnd;

  if pUseTexture then
    FEngine.TextureManager.BindTexture(TextFG);  //2.Textur

  perc:= Value/MaxVal;
  if perc>1 then perc:= 1;
  if Horizontal then
    TmpWidth  := TmpWidth * perc
  else
    TmpHeight := TmpHeight * perc;

  // Alpha Wert setzen
  if FUseAlpha then
  begin
    RGBAf(ColorFG, FAlpha);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glEnable(GL_BLEND);
  end
  else
    RGBf(ColorFG);

  glBegin(GL_QUADS);
    glTexCoord2f(0, 0); glVertex3f(0,        0, Z - MaxDepth + 2);
    glTexCoord2f(1, 0); glVertex3f(TmpWidth, 0, Z - MaxDepth + 2);
    glTexCoord2f(1, 1); glVertex3f(TmpWidth,TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f(0, 1); glVertex3f(0,       TmpHeight, Z - MaxDepth + 2);
  glEnd;
end;

{ TListBox }

constructor TListBox.Create(pEngine: TEngine);
begin
  inherited;
  Items:= TStringList.Create;
  ItemIndex:= 0;
  Font:= 'default';
  TextColor:= clGreen;
  SelectedColor:= clLime;
  Border:= 3;
end;

destructor TListBox.Destroy;
begin
  Items.Free;
  inherited;
end;


procedure TListBox.BindTexture;
begin
  FEngine.TextureManager.BindTexture(Texture);
end;

procedure TListBox.DrawInternal(pUseTexture: Boolean);
var
  i:integer;
  TmpWidth, TmpHeight,
  h,x,y: single;
begin
  // Das eigentlich "Zeichnen"
  TmpWidth  := (Size.X * (FEngine.Form.ClientWidth  / FEngine.Width))  / 2;
  TmpHeight := (Size.Y * (FEngine.Form.ClientHeight / FEngine.Height)) / 2;

  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

  glBegin(GL_QUADS);
    glTexCoord2f(0,      0);    glVertex3f(-TmpWidth, -TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f(TileX, 0);     glVertex3f( TmpWidth, -TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f(TileX, TileY); glVertex3f( TmpWidth,  TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f(0,     TileY); glVertex3f(-TmpWidth,  TmpHeight, Z - MaxDepth + 2);
  glEnd;

  h:= FEngine.FontManager.GeneralTextHeight(Font);
  x:= Border - TmpWidth;
  y:= Border + h*LineSpacing - TmpHeight;
  for i := 0 to Items.count-1 do begin
    if i=ItemIndex then
      FEngine.FontManager.BindFont(Font,SelectedColor)
    else
      FEngine.FontManager.BindFont(Font,TextColor);
    FEngine.TextOut(x, y + i*h*LineSpacing, Items[i]);
  end;
end;

procedure TListBox.MoveSelection(Delta: integer);
begin
  inc(ItemIndex,delta);
  if ItemIndex<0 then inc(ItemIndex,Items.Count);
  if ItemIndex>Items.Count-1 then dec(ItemIndex,Items.Count);
end;

function TListBox.SelectedItem: string;
begin
  Result:= Items[Itemindex];
end;

end.
