// Unit Objects
(**************************************************

- Enthlt: TSimpleObject
           TAnimObject
           TTileset
           TTile

**************************************************)

unit F2D_Objects;

interface

uses
  // Bentigte Units
  Graphics, IniFiles, dglOpenGL, glBitmap,
  Windows,
  F2D_Vectors,
  F2D_Const,
  F2D_Core,
  F2D_Colors,
  F2D_TextureManager;

type
  // Klasse mit den immer bentigten Eigenschaften wie Position und Gre
  TGraphicObject = class
  public
    Position, Size,
    Ofs: TVector2D;   {--SH 20061110}
    // Z gibt die Tiefe an, wobei 0 ganz hinten ist
    Z: Integer;
    constructor Create;
    function BoundsRect:TRect;
  end;

  TBasicObject = class(TGraphicObject)
  protected
    FEngine: TEngine;              // Pointer auf die Engine
    FAngle: Single;                // Ausrichtung in Grad (0 - 360)
    FVisible: Boolean;             // Sichtbar oder Unsichtbar
    FColor: TColor;                // Farbe die benutzt wird (allein oder gem.)
    FUseAlpha: Boolean;            // Alpha fr Blending
    FAlpha: Integer;               // Alpha-Wert (0 - 255)
    FAlphaTest: Boolean;           // AlphaTest - fr Bilder mit Alpha-Kanal
    FAlphaTolerance: Single;       // Toleranz fr den AlphaTest (Default: 0.1)
    procedure BindTexture; virtual;
    procedure DrawInternal(pUseTexture: Boolean); virtual;
    procedure SetMatrix; virtual;
  public
    constructor Create(pEngine: TEngine); virtual;
    // Prozeduren zum setzen der Angle und Alpha Werte
    procedure SetAngle(const Value: Single);
    procedure SetAlpha(const Value: Integer);
    procedure SetAlphaTolerance(const Value: Single);

    // Die verfgbaren Werte als propertys
    property Angle: Single read FAngle write SetAngle;
    property Visible: Boolean read FVisible write FVisible;
    property Color: TColor read FColor write FColor;
    property UseAlpha: Boolean read FUsealpha write FUseAlpha;
    property Alpha: Integer read FAlpha write SetAlpha;
    property AlphaTest: Boolean read FAlphaTest write FAlphaTest;
    property AlphaTolerance: Single read FAlphaTolerance write SetAlphaTolerance;

    procedure Draw(pUseTexture: Boolean);
  end;

  // Einfaches Objekt, welches durch eine Textur dargestellt wird
  TSimpleObject = class(TBasicObject)
  private
    FTexture: String;              // Der Name der benutzten Textur
    FTileX,
    FTileY: single;
    procedure SetTileX(const Value: single);
    procedure SetTileY(const Value: single);
  protected
    procedure BindTexture;   override;
    procedure DrawInternal(pUseTexture: Boolean); override;
  public
    constructor Create(pEngine: TEngine);   override;

    // Generiert und zeichnet einen Schatten fr das Objekt pScaleFaktor
    // gibt die Vergrerung / Verkleinerung an pShadowOffset die Verschiebung
    procedure DrawShadow(pShadowOffset: TVector2D; pScaleFactor: Single);

    property Texture: String read FTexture write FTexture;
    property TileX: single read FTileX write SetTileX;
    property TileY: single read FTileY write SetTileY;
  end;

  // TAnimation beschreibt eine einzige Animation
  TAnimation = class
    constructor Create;
    destructor Destroy; override;
    procedure Free;
  private
    Position: single;
    CoordX, CoordY: Integer;
    FFinished: boolean;
  public

    Texture: String;

    // Hier muss die Anzahl der Bilder auf der X / Y Achse angegeben werden
    AnimCountX, AnimCountY: Integer;
    // Duration gibt die Dauer der Animation in ms an
    Duration: Integer;

    procedure Reset;
    procedure play(pLoop: boolean; pTimestep: single);
    property Finished: boolean read FFinished;
  end;

  // Das Objekt AnimObj kann mehrere Animation speichern und darstellen
  TAnimObj = class(TBasicObject)

    // Hier werden die Animationen gespeichert
    Animations: THashedStringList;
    Loop:boolean;
    Animation: String;
  protected
    procedure BindTexture;  override;
    procedure DrawInternal(pUseTexture: Boolean); override;
  public
    constructor Create(pEngine: TEngine); override;
    destructor Destroy; override;

    // Hier kann eine Animation hinzugefgt werden. Verwendet werden kann sie
    // dann noch nicht, da AnimCount und Duration Werte gesetzt werden mssen
    procedure AddAnimation(pTexturename, pAnimationName: string);

    // Mit "with GetAnimation(ANIMATIONNAME) do" knnen der Animation AnimCount
    // und DurationWerte bergeben werden somit wird einem erspart selbst
    // eine Klasse instanzieren zu mssen
    function GetAnimation(pAnimationName: String): TAnimation;

    //Setzt die Animation zurck
    procedure ResetAnimation(pAnimationName: string);

    // Spielt die Animation - loop gibt an, ob die Animation neu gestartet werden soll
    procedure Play(pAnimationName: String; pLoop: Boolean);
  end;

  // TileSet ist wie eine Animation aufgebaut, hier knnen mehrere Texturen zu
  // einer zusammengefasst werden. Die Textur wird ber Texture angegeben
  // TileCount gibt die Anzahl der Bilder auf der X / Y Achse an
  TTileSet = record
    Texture: String;
    TileCountX, TileCountY: Integer;
  end;

  // Tile greift auf ein Bild aus einem Tileset zu und stellt es dar
  TTile = class(TBasicObject)
  protected
    TileSet:TTileSet;
    procedure BindTexture;   override;
    procedure DrawInternal(pUseTexture: Boolean); override;
  public
    // Tile gibt die Position des Tiles auf der X / Y Achse an
    // Dabei ist das erste Element das mit dem Index 0
    TileX, TileY: Integer;

    constructor Create(pEngine: TEngine); override;
    
    // Zeichnet das Tile, hierbei muss noch das TileSet angegeben werden
    procedure Draw(pTileSet: TTileSet);
  end;

implementation

constructor TGraphicObject.Create;
begin
  inherited Create;

  // Werte auf 0 setzen
  Position := To_V2D(0, 0);
  Size := To_V2D(0, 0);
  Ofs := To_V2D(0, 0);       {--SH 20061110}
  Z := 0;
end;

function TGraphicObject.BoundsRect: TRect;
begin
  result.Left:= trunc((Position.X-Ofs.X) - Size.X/2);
  result.Top:= trunc((Position.Y-Ofs.Y) - Size.Y/2);
  result.Right:= trunc((Position.X-Ofs.X) + Size.X/2);
  result.Bottom:= trunc((Position.Y-Ofs.Y) + Size.Y/2);
end;

constructor TBasicObject.Create(pEngine: TEngine);
begin
  inherited Create;

  // Werte Initialisieren
  FAngle := 0;
  FVisible := True;

  // Farbe auf Wei setzen -> Keine Frbung
  FColor := clWhite;
  FUseAlpha := False;
  FAlpha := 255;
  FAlphaTest := False;

  FEngine := pEngine;

  // AlphaToleranz default-Wert
  FAlphaTolerance := 0.8;
end;

procedure TBasicObject.SetAngle(const Value: Single);
begin
  // Nur korrekte Werte fr den Winkel setzen setzen
  if (Value > 360) then
  begin
    FAngle := Value - 360;
    Exit;
  end;
  if Value < 0 then
  begin
    FAngle := 360 + Value;
    Exit;
  end;

  FAngle := Value;
end;

procedure TBasicObject.SetAlpha(const Value: Integer);
begin
  // Nur Zahlen von 0 - 255 zulassen
  if (Value > 255) OR (Value < 0) then
  begin
    FAlpha := 0;
    Exit;
  end;

  FAlpha := Value;
end;

procedure TBasicObject.SetAlphaTolerance(const Value: Single);
begin
  // AlphaToleranz liegt zwischen 0 und 1
  // Dabei sollte es vermieden werden 1 zu whlen
  if Value < 0 then
  begin
    FAlphaTolerance := 0;
    Exit;
  end;
  if Value > 1 then
  begin
    FAlphaTolerance := 1;
    Exit;
  end;

  FAlphaTolerance := Value;
end;

procedure TBasicObject.BindTexture;
begin
end;

procedure TBasicObject.DrawInternal(pUseTexture: Boolean);
begin
end;

procedure TBasicObject.SetMatrix;
begin
  // Matrix auf die Zeichenposition setzen
  {--SH 20061110 BEGIN}
  glTranslatef((Position.X-Ofs.X) * (FEngine.Form.ClientWidth / FEngine.Width),
               (Position.Y-Ofs.Y) * (FEngine.Form.ClientHeight / FEngine.Height), Z);
  {--SH 20061110 END}
  // Drehen
  glRotatef(FAngle, 0, 0, 1);
end;

procedure TBasicObject.Draw(pUseTexture: Boolean);
begin
  // Nichts zeichnen
  if not Visible then
    Exit;

  // Matrix auf den Matrix Stack schieben
  glPushMatrix;

  SetMatrix;

  if not pUseTexture then
    glDisable(GL_TEXTURE_2D)
  else
    BindTexture;

  // Alpha Wert setzen
  if FUseAlpha then
  begin
    RGBAf(FColor, FAlpha);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
  end
  else
    RGBf(FColor);

  // AlphaTest
  if FAlphaTest then
  begin
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GREATER, FAlphaTolerance);
  end;

  DrawInternal(pUseTexture);

  // Alpha wieder deaktivieren
  if FUseAlpha then
    glDisable(GL_BLEND);

  // Alphatest wieder deaktivieren
  if FAlphaTest then
    glDisable(GL_ALPHA_TEST);

  // Texture wieder aktivieren, falls vorher deaktiviert
  if not pUseTexture then
    glEnable(GL_TEXTURE_2D);
  // Texture wieder rauswerfen
  FEngine.TextureManager.UnBindTexture;

  // Matrix zurcksetzen
  glPopMatrix;
end;



constructor TSimpleObject.Create(pEngine: TEngine);
begin
  inherited Create(pEngine);
  FTexture := '';
  FTileX := 1;
  FTileY := 1;
end;

procedure TSimpleObject.SetTileX(const Value: single);
begin
  {--SH 20061110 removed check to enable full scaling}
  FTileX := Value;
end;

procedure TSimpleObject.SetTileY(const Value: single);
begin
  {--SH 20061110 removed check to enable full scaling}
  FTileY := Value;
end;

procedure TSimpleObject.BindTexture;
begin
  FEngine.TextureManager.BindTexture(FTexture);
end;

procedure TSimpleObject.DrawInternal(pUseTexture: Boolean);
var
  TmpWidth, TmpHeight: Extended;
begin
  // Das eigentlich "Zeichnen"
  TmpWidth  := (Size.X * (FEngine.Form.ClientWidth  / FEngine.Width))  / 2;
  TmpHeight := (Size.Y * (FEngine.Form.ClientHeight / FEngine.Height)) / 2;

  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

  glBegin(GL_QUADS);
    glTexCoord2f(0,      0);      glVertex3f(-TmpWidth, -TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f(FTileX, 0);      glVertex3f( TmpWidth, -TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f(FTileX, FTileY); glVertex3f( TmpWidth,  TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f(0,      FTileY); glVertex3f(-TmpWidth,  TmpHeight, Z - MaxDepth + 2);
  glEnd;
end;

procedure TSimpleObject.DrawShadow(pShadowOffset: TVector2D; pScaleFactor: Single);
var
  TmpWidth, TmpHeight: Extended;
begin
  if not FVisible then
    Exit;

  glPushMatrix;
    {--SH 20061110 BEGIN}
    glTranslatef((Position.X-Ofs.X) * (FEngine.Form.ClientWidth / FEngine.Width),
                 (Position.Y-Ofs.Y) * (FEngine.Form.ClientHeight / FEngine.Height), 0);
    {--SH 20061110 END}
    glRotatef(FAngle, 0, 0, 1);
    glScalef(pScaleFactor, pScaleFactor, pScaleFactor);
    FEngine.TextureManager.BindTexture(FTexture);

    // AlphaTest aktivieren
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GREATER, FAlphaTolerance);

    // Farbe auf Schwarz setzen
    RGBf(clBlack);

    // Zeichnen
    TmpWidth  := (Size.X * (FEngine.Form.ClientWidth  / FEngine.Width))  / 2 + pShadowOffset.X;
    TmpHeight := (Size.Y * (FEngine.Form.ClientHeight / FEngine.Height)) / 2 + pShadowOffset.Y;

    glBegin(GL_QUADS);
      glTexCoord2f(0,0); glVertex3f(-TmpWidth, -TmpHeight, Z - MaxDepth + 1);
      glTexCoord2f(1,0); glVertex3f( TmpWidth, -TmpHeight, Z - MaxDepth + 1);
      glTexCoord2f(1,1); glVertex3f( TmpWidth,  TmpHeight, Z - MaxDepth + 1);
      glTexCoord2f(0,1); glVertex3f(-TmpWidth,  TmpHeight, Z - MaxDepth + 1);
    glEnd;

    // AlphaTest deaktivieren
    glDisable(GL_ALPHA_TEST);

  glPopMatrix;
end;

constructor TAnimation.Create;
begin
  inherited Create;

  // Initialisieren
  CoordX := 0;
  CoordY := 0;
  Position := 0;
  FFinished := false;
end;

destructor TAnimation.Destroy;
begin
  inherited Destroy;
end;

procedure TAnimation.Free;
begin
  if Self <> nil then
    Destroy;
end;

procedure TAnimation.Reset;
begin
  // Entspricht der Initialisierung
  CoordX := 0;
  CoordY := 0;
  Position := 0;
  FFinished := false;
end;

procedure TAnimation.play(pLoop: boolean; pTimestep: single);
begin
  // Das nchste Bild auswhlen
  if (CoordX > -1) and (Coordy > -1) and
  (Coordx <= AnimCountX) and (CoordY <= AnimCountY) then
    Position := Position + 1000 * pTimestep;

  if (Position >= ((Duration) / (AnimCountX * AnimCountY))) then
  begin
    Position := 0;
    inc(CoordX);

    if CoordX > (AnimCountX - 1) then
    begin
      CoordX := 0;
      inc(CoordY);
    end;

    // Wiederholen ?
    if (CoordY > AnimCountY - 1)  then
      if ploop then
      begin
        CoordX := 0;
        CoordY := 0;
      end else
      begin
        Coordx := -1;
        Coordy := -1;
        FFinished := true;
      end;
  end;
end;

constructor TAnimObj.Create(pEngine: TEngine);
begin
  inherited Create(pEngine);

  Animations := THashedStringList.Create;
end;

destructor TAnimObj.Destroy;
var
  I: Integer;
begin
  // Animationen freigeben
  for I := 0 to Animations.Count-1 do
    Animations.Objects[I].Free;
  Animations.Free;
  
  inherited Destroy;
end;

procedure TAnimObj.AddAnimation(pTextureName, pAnimationName: String);
var a:TAnimation;
begin
  a := TAnimation.Create;           // Animation erstellen
  a.Texture := pTextureName;        // Textur zuweisen
  Animations.AddObject(pAnimationName,a);                 // Der Liste hinzufgen
end;

function TAnimObj.GetAnimation(pAnimationName: String): TAnimation;
begin
  // Gibt die Animation zurck
  Result := TAnimation(Animations.Objects[animations.IndexOf(pAnimationName)]);
end;

procedure TAnimObj.ResetAnimation(pAnimationName: string);
begin
  TAnimation(Animations.Objects[animations.IndexOf(pAnimationName)]).reset;
end;

procedure TAnimObj.BindTexture;
begin
  // Textur auswhlen
  FEngine.TextureManager.BindTexture(
     TAnimation(Animations.Objects[animations.IndexOf(Animation)]).Texture);
end;

procedure TAnimObj.DrawInternal(pUseTexture: Boolean);
var
  index: integer;
  tmpwidth,
  tmpheight: single;
begin
  index := animations.IndexOf(Animation);

  TmpWidth  := Size.X * (FEngine.Form.ClientWidth  / FEngine.Width)  / 2;
  TmpHeight := Size.Y * (FEngine.Form.ClientHeight / FEngine.Height) / 2;

  // Zeichnen
  glBegin(GL_QUADS);
  with TAnimation(Animations.Objects[Index]) do
  begin
    glTexCoord2f( CoordX      * (1 / AnimCountX),  CoordY      * (1 / AnimCountY));
    glVertex3f(-TmpWidth, -TmpHeight, Z - MaxDepth + 2);

    glTexCoord2f((CoordX + 1) * (1 / AnimCountX),  CoordY      * (1 / AnimCountY));
    glVertex3f( TmpWidth, -TmpHeight, Z - MaxDepth + 2);

    glTexCoord2f((CoordX + 1) * (1 / AnimCountX), (CoordY + 1) * (1 / AnimCountY));
    glVertex3f( TmpWidth,  TmpHeight, Z - MaxDepth + 2);

    glTexCoord2f( Coordx      * (1 / AnimCountX), (CoordY + 1) * (1 / AnimCountY));
    glVertex3f(-TmpWidth,  TmpHeight, Z - MaxDepth + 2);
  end;
  glEnd;

  TAnimation(Animations.Objects[Index]).play(Loop, FEngine.Timestep);
end;

procedure TAnimObj.Play(pAnimationName: String; pLoop: Boolean);
begin
  Loop:= pLoop;
  Animation:= pAnimationName;
  Draw(true);
end;

constructor TTile.Create(pEngine: TEngine);
begin
  inherited Create(pEngine);
  TileX := 0;
  TileY := 0;
end;

procedure TTile.BindTexture;
begin
  // TileSet Textur auswhlen
  FEngine.TextureManager.BindTexture(TileSet.Texture);
end;

procedure TTile.DrawInternal(pUseTexture: Boolean);
var
  TmpX, TmpY, TmpWidth, TmpHeight: Extended;
begin
  // Zeichnen
  TmpX := 1 / TileSet.TileCountX;
  TmpY := 1 / TileSet.TileCountY;
  TmpWidth  := Size.X * (FEngine.Form.ClientWidth  / FEngine.Width)  / 2;
  TmpHeight := Size.Y * (FEngine.Form.ClientHeight / FEngine.Height) / 2;

  glBegin(GL_QUADS);
    glTexCoord2f( TileX *      TmpX, TileY *       TmpY); glVertex3f(-TmpWidth, -TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f((TileX + 1) * TmpX, TileY *       TmpY); glVertex3f( TmpWidth, -TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f((TileX + 1) * TmpX, (TileY + 1) * TmpY); glVertex3f( TmpWidth,  TmpHeight, Z - MaxDepth + 2);
    glTexCoord2f( TileX *      TmpX, (TileY + 1) * TmpY); glVertex3f(-TmpWidth,  TmpHeight, Z - MaxDepth + 2);
  glEnd;
end;

procedure TTile.Draw(pTileSet: TTileSet);
begin
  TileSet:= pTileSet;
  inherited Draw(true);
end;

end.
